#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

import struct
import threading
import time
import traceback
import warnings
from typing import Optional, Union, Tuple, Any
from oser.types import OserElementType

import serial  # type: ignore
import socket
import ssl
from binascii import hexlify, unhexlify

from htf.modbus.protocol import (
    SerialRTUModbusRequest,
    SerialRTUModbusResponse,
    SerialASCIIModbusResponse,
    SerialASCIIModbusRequest,
    TCPModbusRequest,
)


def _create_ssl_client_context() -> ssl.SSLContext:  # pragma: no cover
    return ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)


def _create_ssl_server_context() -> ssl.SSLContext:  # pragma: no cover
    return ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)


class ModbusBaseInterface:
    pass


class ModbusSerialRTUClientInterface(ModbusBaseInterface):
    def __init__(
        self,
        com_port: Optional[str] = None,
        baud: int = 9600,
        bytesize: int = serial.EIGHTBITS,
        parity: str = serial.PARITY_EVEN,
        stopbits: int = serial.STOPBITS_ONE,
        rtscts: bool = False,
        timeout: Union[int, float] = 1.0,
        response_parser: OserElementType = SerialRTUModbusResponse,
    ) -> None:
        """
        Args:
            com_port: the com port to be used
            baud: the baud rate
            bytesize: number of bits in a byte
            parity: serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits: number of stopbits (1, 1.5 or 2)
            rtscts: if True use flow control
            timeout: the timeout in seconds
            response_parser: the parser to parse responses
        """
        self.baud = baud
        if parity == serial.PARITY_NONE and stopbits != serial.STOPBITS_TWO:
            warnings.warn(Warning("Two stop bits are required when using no parity!"))
        self.serial = serial.serial_for_url(
            com_port,
            baud,
            bytesize=bytesize,
            parity=parity,
            stopbits=stopbits,
            rtscts=rtscts,
            timeout=timeout,
            write_timeout=timeout,
        )
        self.last_write = time.time()
        self._response_parser = response_parser

    def read(self) -> Tuple[bytes, None]:
        """
        Read Modbus serial RTU responses from server.

        Returns:
            a tuple containing buffer, ``None``
        """
        buffer = b""
        decoder = self._response_parser()
        read = 0

        while read <= 256:
            # read byte wise and try to decode the request
            byte = self.serial.read(1)

            if not byte:
                return buffer, None

            read += len(byte)

            buffer += byte

            try:
                decoder.decode(buffer)
                return buffer, None
            except Exception:
                pass

        return buffer, None

    def write(self, adu: SerialRTUModbusRequest) -> None:
        """
        Write ``adu``.

        Args:
            adu: serializable RTU adu
        """
        self.serial.flushInput()
        try:
            current = time.time()
            time_diff = current - (self.last_write + (3.5 * 11) / self.baud)  # wait at least 3.5 char times
            if time_diff < 0.0:
                time.sleep(min(0.0, -time_diff))
            data = adu.encode()
            self.serial.write(data)
            self.last_write = current
        except Exception:  # pragma: no cover
            traceback.print_exc()

    def close(self) -> None:
        """
        Close serial port.
        """
        self.serial.close()


class ModbusSerialRTUServerInterface(ModbusSerialRTUClientInterface):
    def __init__(
        self,
        com_port: Optional[str] = None,
        baud: int = 9600,
        bytesize: int = serial.EIGHTBITS,
        parity: str = serial.PARITY_EVEN,
        stopbits: int = serial.STOPBITS_ONE,
        rtscts: bool = False,
        timeout: Union[int, float] = 1.0,
        response_parser: OserElementType = SerialRTUModbusRequest,
    ) -> None:
        """
        Args:
            com_port: the com port to be used
            baud: the baud rate
            bytesize: number of bits in a byte
            parity: serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits: number of stopbits (1, 1.5 or 2)
            rtscts: if True use flow control
            timeout: the timeout in seconds
            response_parser: the parser to parse responses
        """
        super(ModbusSerialRTUServerInterface, self).__init__(
            com_port=com_port,
            baud=baud,
            bytesize=bytesize,
            parity=parity,
            stopbits=stopbits,
            rtscts=rtscts,
            timeout=timeout,
            response_parser=response_parser,
        )

    def read(self) -> Tuple[bytes, None]:
        """
        Read Modbus serial RTU requests from client.

        Returns:
            a tuple containing buffer, ``None``
        """
        return super(ModbusSerialRTUServerInterface, self).read()


class ModbusSerialASCIIClientInterface(ModbusBaseInterface):
    def __init__(
        self,
        com_port: Optional[str] = None,
        baud: int = 9600,
        bytesize: int = serial.EIGHTBITS,
        parity: str = serial.PARITY_EVEN,
        stopbits: int = serial.STOPBITS_ONE,
        rtscts: bool = False,
        timeout: Union[int, float] = 1.0,
        response_parser: OserElementType = SerialASCIIModbusResponse,
    ) -> None:
        """
        Args:
            com_port: the com port to be used
            baud: the baud rate
            bytesize: number of bits in a byte
            parity: serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits: number of stopbits (1, 1.5 or 2)
            rtscts: if True use flow control
            timeout: the timeout in seconds
            response_parser: the parser to parse responses
        """
        self.ascii_delimiter = b"\x0A"
        self.baud = baud
        if parity == serial.PARITY_NONE and stopbits != serial.STOPBITS_TWO:
            warnings.warn(Warning("Two stop bits are required when using no parity!"))
        self.serial = serial.serial_for_url(
            com_port,
            baud,
            bytesize=bytesize,
            parity=parity,
            stopbits=stopbits,
            rtscts=rtscts,
            timeout=timeout,
            write_timeout=timeout,
        )
        self._response_parser = response_parser

    def change_ascii_delimiter(self, delimiter: bytes) -> None:
        """
        Change LF character to ``delimiter``.

        Args:
            delimiter: bytes of length 1 containing the new ASCII delimiter
        """
        assert isinstance(delimiter, (bytes, bytearray)) and len(delimiter) == 1
        self.ascii_delimiter = bytes(delimiter)

    def _bytes_to_ascii(self, encoded: bytes) -> bytes:  # pragma: no cover
        ret = b"\x3A"
        for byte in hexlify(encoded):
            ret += struct.Struct(">B").pack(byte)
        return ret + b"\x0D" + self.ascii_delimiter

    def _ascii_to_bytes(self, ascii: bytes) -> bytes:
        return unhexlify(ascii[1:-2])

    def read(self) -> Tuple[bytes, None]:
        """
        Read Modbus serial ASCII responses from server.
        Read until line feed character.

        Returns:
            a tuple containing buffer, ``None``
        """

        buffer = b""

        decoder = self._response_parser()
        read = 0
        while read <= 517:
            byte = self.serial.read(1)
            if not byte:
                return b"", None

            read += len(byte)

            if byte == b"\x3A" or buffer.startswith(b"\x3A"):  # pragma: no cover
                # frame start received
                buffer += byte
            else:  # pragma: no cover
                continue

            if byte == self.ascii_delimiter:
                # line feed received
                try:
                    bytes_buffer = self._ascii_to_bytes(buffer)
                    decoder.decode(bytes_buffer)
                    return bytes_buffer, None
                except Exception:
                    pass
                return self._ascii_to_bytes(buffer), None

        return self._ascii_to_bytes(buffer), None

    def write(self, adu: SerialASCIIModbusRequest) -> None:
        """
        Write ``adu``.

        Args:
            adu: serializable RTU adu
        """
        self.serial.flushInput()
        data = self._bytes_to_ascii(adu.encode())
        self.serial.write(data)

    def close(self) -> None:
        """
        Close serial port.
        """
        self.serial.close()


class ModbusSerialASCIIServerInterface(ModbusSerialASCIIClientInterface):
    def __init__(
        self,
        com_port: Optional[str] = None,
        baud: int = 9600,
        bytesize: int = serial.EIGHTBITS,
        parity: str = serial.PARITY_EVEN,
        stopbits: int = serial.STOPBITS_ONE,
        rtscts: bool = False,
        timeout: Union[int, float] = 1.0,
        response_parser: OserElementType = SerialASCIIModbusRequest,
    ) -> None:
        """
        Args:
            com_port: the com port to be used
            baud: the baud rate
            bytesize: number of bits in a byte
            parity: serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits: number of stopbits (1, 1.5 or 2)
            rtscts: if True use flow control
            timeout: the timeout in seconds
            response_parser: the parser to parse responses
        """
        super(ModbusSerialASCIIServerInterface, self).__init__(
            com_port=com_port,
            baud=baud,
            bytesize=bytesize,
            parity=parity,
            stopbits=stopbits,
            rtscts=rtscts,
            timeout=timeout,
            response_parser=response_parser,
        )

    def read(self) -> Tuple[bytes, None]:
        """
        Read serial ASCII Modbus requests from client.

        Returns:
            a tuple containing buffer, ``None``
        """
        return super(ModbusSerialASCIIServerInterface, self).read()


class ModbusTCPClientInterface(ModbusBaseInterface):
    _socket_type = socket.SOCK_STREAM

    def __init__(
        self,
        address: str,
        port: int = 502,
        timeout: Union[int, float] = 1.0,
        connect: bool = True,
    ):
        """
        Args:
            address: the ip-address to connect to
            port: the port
            timeout: the timeout in seconds
            connect: if set to ``True`` the socket connects automatically
        """
        self.socket: Optional[socket.socket] = None
        self.address = address
        self.port = port
        self.timeout = timeout
        self.socket_lock = threading.Lock()
        if connect:
            self.open()

    def read(self) -> Tuple[Optional[bytes], None]:
        """
        Read a TCP request from TCP server.

        Returns:
            a tuple containing buffer, ``None``
        """
        try:
            return self.socket.recv(300), None  # type: ignore
        except Exception:
            return None, None

    def write(self, adu: TCPModbusRequest) -> None:
        """
        Write ``adu``.

        Args:
            adu: serializable TCP adu
        """
        self.socket.sendall(adu.encode())  # type: ignore

    def open(self) -> None:
        """
        Open a socket.
        """
        try:
            with self.socket_lock:
                self.socket = socket.socket(type=self._socket_type)
                self.timeout = self.timeout
                self.socket.setblocking(True)
                self.socket.settimeout(self.timeout)

                self.socket.connect((self.address, self.port))
        except Exception:
            self.close()
            raise

    def close(self) -> None:
        """
        Close TCP connection.
        """
        with self.socket_lock:
            self.socket.close()  # type: ignore

    def reconnect(self) -> None:
        """
        Close socket and create a new one.
        """
        self.close()
        self.open()


class ModbusTLSClientInterface(ModbusTCPClientInterface):
    def __init__(
        self,
        address: str,
        port: int = 502,
        timeout: Union[int, float] = 1.0,
        verify: bool = True,
        connect: bool = True,
        cafile: Optional[str] = None,
        capath: Optional[str] = None,
        cadata: Optional[str] = None,
    ) -> None:
        """
        Args:
            address: the ip-address to connect to
            port: the port
            timeout: the timeout in seconds
            verify: set to ``False`` to disable TLS certificate verification
            connect: if set to ``True`` the socket connects automatically
            cafile: path to a file of concatenated CA certificates in PEM format
            capath: path to a directory containing several CA certificates in PEM format
            cadata: either an ASCII string of one or more PEM-encoded certificates or a bytes-like
                object of DER-encoded certificates
        """

        self.socket = None
        self.address = address
        self.port = port
        self.timeout = timeout
        self.cafile = cafile
        self.capath = capath
        self.cadata = cadata
        self.verify = verify
        self.socket_lock = threading.Lock()

        if connect:
            self.open()

    def open(self) -> None:
        """
        Open a socket.
        """

        try:
            with self.socket_lock:
                context = _create_ssl_client_context()
                context.load_verify_locations(cafile=self.cafile, capath=self.capath, cadata=self.cadata)

                if self.verify is False:
                    # insecure mode
                    context.check_hostname = False
                    context.verify_mode = ssl.CERT_NONE
                else:
                    context.verify_mode = ssl.CERT_REQUIRED
                    context.check_hostname = True

                sock = socket.socket(type=self._socket_type)
                self.timeout = self.timeout
                sock.settimeout(self.timeout)
                self.socket = context.wrap_socket(sock, server_hostname=self.address)

                self.socket.connect((self.address, self.port))
        except Exception:
            self.close()
            raise


class ModbusUDPClientInterface(ModbusTCPClientInterface):
    _socket_type = socket.SOCK_DGRAM

    def __init__(
        self,
        address: str,
        port: int = 502,
        timeout: Union[int, float] = 1.0,
    ) -> None:
        """
        Args:
            address: the ip-address to connect to
            port: the port
            timeout: the timeout in seconds
        """
        self.socket = socket.socket(type=self._socket_type)
        self.socket.setblocking(True)
        self.timeout = timeout
        self.socket_lock = threading.Lock()
        self.socket.settimeout(self.timeout)
        self.address = address
        self.port = port

    def read(self) -> Tuple[Optional[bytes], None]:
        """
        Read a TLS request from TLS server.

        Returns:
            a tuple containing buffer, ``None``
        """
        try:
            return self.socket.recvfrom(300)  # type: ignore
        except Exception:
            return None, None

    def write(self, adu: TCPModbusRequest) -> None:
        """
        Write ``adu``.

        Args:
            adu: serializable TCP adu
        """
        self.socket.sendto(adu.encode(), (self.address, self.port))  # type: ignore


class ModbusTCPServerInterface(ModbusBaseInterface):
    _socket_type = socket.SOCK_STREAM

    def __init__(
        self,
        address: str,
        port: int = 502,
        timeout: Union[int, float] = 1.0,
    ):
        """
        Args:
            address: the ip-address to listen on
            port: the port to listen on
            timeout: the timeout in seconds
        """
        self.socket = socket.socket(type=self._socket_type)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.timeout = timeout
        self.socket.settimeout(self.timeout)
        self.socket.bind((address, port))
        self.socket.listen(100)

    def accept(self) -> Tuple[socket.socket, Any]:
        """
        Accept a new connection.

        Returns:
            a tuple containing connection, address
        """
        conn, address = self.socket.accept()
        return conn, address

    def close(self) -> None:
        """
        Close TCP socket.
        """
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        self.socket.close()

    def close_connection(self, connection: socket.socket) -> None:
        """
        Close TCP connection.
        """
        try:
            connection.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        connection.close()

    def read(self, connection: socket.socket) -> bytes:
        """
        Read a TCP request from TLS client.

        Returns:
            a tuple containing buffer, ``None``
        """
        connection.settimeout(self.timeout)
        return connection.recv(300)

    def write(self, connection: socket.socket, adu: TCPModbusRequest) -> None:
        """
        Write ``adu``.

        Args:
            adu: serializable TCP adu
        """
        connection.sendall(adu.encode())


class ModbusTLSServerInterface(ModbusTCPServerInterface):
    def __init__(
        self,
        address: str,
        port: int = 502,
        timeout: Union[int, float] = 1.0,
        certfile: Optional[str] = None,
        keyfile: Optional[str] = None,
        password: Optional[str] = None,
    ) -> None:
        """
        Args:
            address: the ip-address to listen on
            port: the port to listen on
            timeout: the timeout in seconds
            certfile: path to TLS certificate
            keyfile: path to TLS keyfile
            password: TLS password
        """
        context = _create_ssl_server_context()
        context.load_cert_chain(certfile=certfile, keyfile=keyfile, password=password)  # type: ignore

        sock = socket.socket(type=self._socket_type)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.timeout = timeout
        sock.settimeout(self.timeout)
        sock.bind((address, port))
        sock.listen(100)

        self.socket = context.wrap_socket(sock, server_side=True)

    def accept(self) -> Tuple[socket.socket, Any]:
        """
        Accept a new connection.

        Returns:
            a tuple containing connection, address
        """
        conn, address = self.socket.accept()
        return conn, address

    def close(self) -> None:
        """
        Close TLS socket.
        """
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        self.socket.close()


class ModbusUDPServerInterface(ModbusBaseInterface):
    _socket_type = socket.SOCK_DGRAM

    def __init__(
        self,
        address: str,
        port: int = 502,
        timeout: Optional[Tuple[int, float]] = None,
    ):
        """
        Args:
            address: the ip-address to listen on
            port: the port to listen on
            timeout=None: the timeout in seconds
        """
        self.socket = socket.socket(type=self._socket_type)
        self.socket.setblocking(False)
        if timeout is not None:  # pragma: no cover
            self.socket.settimeout(timeout)  # type: ignore
        self.socket.bind((address, port))

    def close(self) -> None:
        """
        Close UDP socket.
        """
        self.socket.close()

    def read(self, bufsize: int = 300) -> Tuple[bytes, Any]:
        """
        Read a request from an UDP client.

        Args:
            bufsize: buffer size

        Returns:
            the received request
        """
        return self.socket.recvfrom(bufsize)

    def write(self, address: str, adu: TCPModbusRequest) -> None:
        """
        Write ``adu``.

        Args:
            address: the target ip-address
            adu: serializable TCP adu
        """
        self.socket.sendto(adu.encode(), address)
