#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

import socket
from typing import Union, Any, Optional, Callable, List, Dict
import oser
from oser.types import OserElementType

import serial  # type: ignore
from htf.modbus.interface import (
    ModbusTCPServerInterface,
    ModbusTLSServerInterface,
    ModbusUDPServerInterface,
    ModbusSerialRTUServerInterface,
    ModbusSerialASCIIServerInterface,
)
from htf.modbus.protocol import (
    SerialRTUModbusRequest,
    SerialRTUModbusResponse,
    TCPModbusRequest,
    TCPModbusResponse,
    SerialASCIIModbusRequest,
    SerialASCIIModbusResponse,
)
from htf.modbus.core import ModbusBaseServer as _ModbusBaseServer
import threading

BYTE_SIZE = 8
MAX_SERIAL_ADU_SIZE = 256
MAX_TCP_ADU_SIZE = 260


class ModbusBaseServer(_ModbusBaseServer):
    def __init__(self) -> None:
        """
        Modbus Base Server is an abstract base class.
        """
        self.listen_only: bool = False
        self.device_address: int = 0
        self.diagnostics: bytes = b"\x00" * 2
        self.bus_message_count: int = 0
        self.bus_communication_error_count: int = 0
        self.bus_exception_error_count: int = 0
        self.slave_message_count: int = 0
        self.slave_no_response_count: int = 0
        self.slave_busy_count: int = 0
        self.bus_character_overrun_count: int = 0
        self.interface: Any = None

    def handle_request(self, data: bytes) -> None:  # pragma: no cover
        """
        Request handler.

        Raises:
            Exception: abstract base class has no implementation.
        """
        raise Exception("Not implemented! Abstract base class.")

    def start(self) -> None:  # pragma: no cover
        """
        Start the server.

        Raises:
            Exception: abstract base class has no implementation.
        """
        raise Exception("Not implemented! Abstract base class.")

    def stop(self) -> None:  # pragma: no cover
        """
        Stop the server.

        Raises:
            Exception: abstract base class has no implementation.
        """
        raise Exception("Not implemented! Abstract base class.")

    def set_listen_only_mode(self, listen_only_mode_enabled: bool = True) -> None:
        """
        Set listen only mode.

        Args:
            listen_only_mode_enabled: ``True`` enables the listen only mode. ``False`` disables it.
        """
        self.listen_only = listen_only_mode_enabled

    def clear_counters(self) -> None:
        """
        Clear all counters.
        """
        self.bus_message_count = 0
        self.bus_communication_error_count = 0
        self.bus_exception_error_count = 0
        self.slave_message_count = 0
        self.slave_no_response_count = 0
        self.slave_busy_count = 0
        self.bus_character_overrun_count = 0

    def clear_diagnostics(self) -> None:
        """
        Clear diagnostics.
        """
        self.diagnostics = b"\x00" * 2

    def close(self) -> None:  # pragma: no cover
        """
        Close interface and stop the server.
        """
        self.interface.close()
        self.stop()


class ModbusSerialRTUServer(ModbusBaseServer):
    _interface_type: Any = ModbusSerialRTUServerInterface

    def __init__(
        self,
        com_port: str,
        baud: int = 9600,
        bytesize: int = serial.EIGHTBITS,
        parity: str = serial.PARITY_EVEN,
        stopbits: int = serial.STOPBITS_ONE,
        rtscts: bool = False,
        device_address: int = 1,
        timeout: Union[int, float] = 1.0,
        request_parser: OserElementType = SerialRTUModbusRequest,
        response_parser: OserElementType = SerialRTUModbusResponse,
        callback: Optional[Callable] = None,
    ) -> None:
        """
        Args:
            com_port: the com port to be used
            baud: the baud rate
            bytesize: number of bits in a byte
            parity: serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits: number of stopbits (1, 1.5 or 2)
            rtscts: if True use flow control
            device_address: the device address to be used (Slave ID)
            timeout: the timeout in seconds
            request_parser: the request parser
            response_parser: the response parser
            callback: a callback used as a request handler
        """
        super(ModbusSerialRTUServer, self).__init__()
        self.device_address = device_address
        self.request_parser = request_parser
        self.response_parser = response_parser
        self.interface = self._interface_type(
            com_port,
            baud=baud,
            bytesize=bytesize,
            parity=parity,
            stopbits=stopbits,
            rtscts=rtscts,
            timeout=timeout,
        )

        self.callback = callback

        self.thread: Optional[threading.Thread] = None
        self.run = False

        self.receive_event_callback: Optional[Callable] = None

    def set_receive_event_callback(self, callback: Optional[Callable]) -> None:
        """
        Set the callback to emit events.

        Args:
            callback: the callback to be called
        """
        self.receive_event_callback = callback

    @staticmethod
    def get_broadcast_enabled_function_codes() -> List[int]:
        """
        Return a list of functions codes that are processed for broadcast requests.
        By default the function codes ``5``, ``6``, ``15`` and ``16`` are supported.
        This method can be overwritten easily.

        Returns:
            a list of int containing the supported function codes
        """
        return [5, 6, 15, 16]

    def handle_request(self, data: bytes) -> None:  # noqa: C901: complexity accepted and well tested
        """
        Request handler.
        A request is handled by calling the callback.

        Args:
            data: a request to be handled
        """
        if len(data) < 3:  # pragma: no cover
            self.bus_communication_error_count += 1
            if self.receive_event_callback is not None:  # pragma: no cover
                self.receive_event_callback(
                    is_communication_error=True,
                    is_broadcast=data[0] == b"\x00",
                    is_listen_only_mode=self.listen_only,
                )
            return

        if len(data) > MAX_SERIAL_ADU_SIZE:
            self.bus_character_overrun_count += 1
            if self.receive_event_callback is not None:  # pragma: no cover
                self.receive_event_callback(
                    is_character_overrun=True,
                    is_broadcast=data[0] == b"\x00",
                    is_listen_only_mode=self.listen_only,
                )

        adu = self.request_parser()

        try:
            adu.decode(data)
            self.bus_message_count += 1
            address = adu.address.get()
            is_broadcast = address == 0
            if is_broadcast or address == self.device_address:
                self.slave_message_count += 1
                process_request = True
                if is_broadcast:  # Broadcast
                    if self.receive_event_callback is not None:  # pragma: no cover
                        self.receive_event_callback(
                            is_broadcast=True,
                        )
                    function_code = adu.pdu.function_code.get()
                    # only enabled functions codes are processed for broadcast requests
                    if function_code not in self.get_broadcast_enabled_function_codes():
                        process_request = False

                if process_request:
                    response = self.callback(adu)  # type: ignore
                    if not self.listen_only and not is_broadcast:
                        response.address.set(self.device_address)  # set device address for response
                        self.interface.write(response)
                    else:
                        self.slave_no_response_count += 1
        except oser.ChecksumMismatchException:
            self.bus_communication_error_count += 1
            if self.receive_event_callback is not None:  # pragma: no cover
                self.receive_event_callback(
                    is_character_overrun=True,
                    is_broadcast=data[0] == b"\x00",
                    is_listen_only_mode=self.listen_only,
                )

    def start(self) -> None:
        """
        Start the server.
        """
        self.run = True
        assert self.thread is None

        def _thread() -> None:
            while self.run:
                try:
                    data, _ = self.interface.read()
                    if data:
                        self.handle_request(data)
                except Exception:
                    pass
                    # traceback.print_exc()

        self.thread = threading.Thread(target=_thread)
        self.thread.daemon = True
        self.thread.start()

    def stop(self) -> None:
        """
        Stop the server.
        """
        self.run = False

    def close(self) -> None:
        """
        Close interface and stop the server.
        """
        self.interface.close()
        self.stop()


class ModbusSerialASCIIServer(ModbusSerialRTUServer):
    _interface_type: Any = ModbusSerialASCIIServerInterface

    def __init__(
        self,
        com_port: str,
        baud: int = 9600,
        bytesize: int = serial.EIGHTBITS,
        parity: str = serial.PARITY_EVEN,
        stopbits: int = serial.STOPBITS_ONE,
        rtscts: bool = False,
        device_address: int = 1,
        timeout: Union[int, float] = 1.0,
        request_parser: OserElementType = SerialASCIIModbusRequest,
        response_parser: OserElementType = SerialASCIIModbusResponse,
        callback: Optional[Callable] = None,
    ) -> None:
        """
        Args:
            com_port: the com port to be used
            baud: the baud rate
            bytesize: number of bits in a byte
            parity: serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits: number of stopbits (1, 1.5 or 2)
            rtscts: if True use flow control
            device_address: the device address to be used (Slave ID)
            timeout: the timeout in seconds
            request_parser: the request parser
            response_parser: the response parser
            callback: a callback used as a request handler
        """
        super(ModbusSerialASCIIServer, self).__init__(
            com_port=com_port,
            baud=baud,
            bytesize=bytesize,
            parity=parity,
            stopbits=stopbits,
            rtscts=rtscts,
            device_address=device_address,
            timeout=timeout,
            request_parser=request_parser,
            response_parser=response_parser,
            callback=callback,
        )

    def change_ascii_delimiter(self, delimiter: bytes) -> None:
        """
        Change ASCII delimiter to ``ascii_delimiter``.

        Args:
            delimiter: bytes of length 1 representing the new lf character
        """
        self.interface.change_ascii_delimiter(delimiter)


class ModbusTCPServer(ModbusBaseServer):
    _interface_type: Any = ModbusTCPServerInterface

    def __init__(
        self,
        bind_address: str,
        port: int = 502,
        timeout: Union[int, float] = 1.0,
        callback: Optional[Callable] = None,
        request_parser: OserElementType = TCPModbusRequest,
        response_parser: OserElementType = TCPModbusResponse,
    ) -> None:
        """
        Args:
            bind_address: the ip-address to listen on
            port: the port to listen on
            timeout: the timeout in seconds
            request_parser: the request parser
            response_parser: the response parser
        """
        super(ModbusTCPServer, self).__init__()
        self._init_parsers_variables_and_callback(request_parser, response_parser, callback)
        self.interface = self._interface_type(bind_address, port=port, timeout=timeout)

    def _init_parsers_variables_and_callback(
        self,
        request_parser: OserElementType,
        response_parser: OserElementType,
        callback: Optional[Callable],
    ) -> None:
        self.request_parser = request_parser
        self.response_parser = response_parser

        if callback is None:  # pragma: no cover
            raise Exception("Callback has to be set!")

        self.callback = callback

        self.connections: Dict[socket.socket, Any] = {}

        self.accept_thread: Optional[threading.Thread] = None
        self.threads: List[threading.Thread] = []
        self.run = False
        self._threads_lock = threading.Lock()
        self._connections_lock = threading.Lock()

    def handle_request(self, connection: socket.socket, data: bytes) -> None:  # type: ignore[override]
        """
        The request handler.
        A request is handled by calling the callback.

        Args:
            connection: the current TCP connection
            data: the request data
        """
        if len(data) > MAX_TCP_ADU_SIZE:
            self.bus_character_overrun_count += 1

        adu = self.request_parser()
        adu.decode(data)
        self.bus_message_count += 1
        self.slave_message_count += 1
        response = self.callback(adu)
        response.transaction_identifier.set(adu.transaction_identifier.get())
        response.protocol_identifier.set(adu.protocol_identifier.get())
        response.length.set(len(response.pdu.encode()) + 1)
        response.unit_identifier.set(adu.unit_identifier.get())
        self.interface.write(connection, response)

    def accept_connections(self) -> None:
        """
        A thread to accept new incoming connections.
        """
        while self.run:
            try:
                conn, address = self.interface.accept()
                with self._connections_lock:
                    self.connections[conn] = address

                def _thread(thread_connection: socket.socket) -> None:
                    while self.run:
                        try:
                            data = self.interface.read(thread_connection)
                            if data:
                                self.handle_request(thread_connection, data)
                            else:
                                # If no data was received the connection is closed by the client.
                                # In case of a timeout an exception is raised instead.
                                self.interface.close_connection(thread_connection)
                                break
                        except Exception:
                            pass
                    self.interface.close_connection(thread_connection)
                    with self._connections_lock:
                        try:
                            del self.connections[thread_connection]
                        except KeyError:  # pragma: no cover
                            pass

                with self._threads_lock:
                    thread = threading.Thread(
                        target=_thread,
                        kwargs=dict(thread_connection=conn),
                    )
                    thread.daemon = True
                    self.threads.append(thread)
                    thread.start()
            except Exception:
                pass

    def start(self) -> None:
        """
        Start the server.
        """
        self.run = True

        assert self.accept_thread is None

        self.accept_thread = threading.Thread(target=self.accept_connections)
        self.accept_thread.daemon = True
        self.accept_thread.start()

    def stop(self) -> None:
        """
        Stop the server.
        """
        self.run = False

        with self._threads_lock:
            for thread in self.threads:
                thread.join(timeout=30)
                assert thread.is_alive() is False

            self.threads = []

        try:
            self.accept_thread.join(timeout=30)  # type: ignore
            assert self.accept_thread.is_alive() is False  # type: ignore
        except AttributeError:  # pragma: no cover
            pass

    def close(self) -> None:
        """
        Close the interface and stop the server.
        """
        self.interface.close()
        self.stop()


class ModbusTLSServer(ModbusTCPServer):
    _interface_type: Any = ModbusTLSServerInterface

    def __init__(
        self,
        bind_address: str,
        port: int = 502,
        timeout: Union[int, float] = 1.0,
        callback: Optional[Callable] = None,
        certfile: Optional[str] = None,
        keyfile: Optional[str] = None,
        password: Optional[str] = None,
        request_parser: OserElementType = TCPModbusRequest,
        response_parser: OserElementType = TCPModbusResponse,
    ) -> None:
        """
        Args:
            bind_address: the ip-address to listen on
            port: the port to listen on
            timeout: the timeout in seconds
            callback: the callback to handle requests
            certfile: path to TLS certificate
            keyfile: path to TLS keyfile
            password: TLS password
            request_parser: the request parser
            response_parser: the response parser
        """
        ModbusBaseServer.__init__(self)
        self._init_parsers_variables_and_callback(request_parser, response_parser, callback)
        self.interface = self._interface_type(
            bind_address,
            port=port,
            timeout=timeout,
            certfile=certfile,
            keyfile=keyfile,
            password=password,
        )


class ModbusUDPServer(ModbusTCPServer):
    _interface_type: Any = ModbusUDPServerInterface

    def __init__(
        self,
        bind_address: str,
        port: int = 502,
        timeout: Union[int, float] = 1.0,
        callback: Optional[Callable] = None,
        request_parser: OserElementType = TCPModbusRequest,
        response_parser: OserElementType = TCPModbusResponse,
    ) -> None:
        """
        Args:
            bind_address: the ip-address to listen on
            port: the port to listen on
            timeout=None: the timeout in seconds
            request_parser: the request parser
            response_parser: the response parser
        """
        super(ModbusUDPServer, self).__init__(
            bind_address=bind_address,
            port=port,
            timeout=timeout,
            callback=callback,
            request_parser=request_parser,
            response_parser=response_parser,
        )

    def accept_connections(self) -> None:
        """
        A thread to accept new incoming connections.
        """
        while self.run:
            try:
                data, address = self.interface.read()

                def _thread(thread_address: Any, thread_data: bytes) -> None:
                    try:
                        self.handle_request(thread_address, thread_data)
                    except Exception:
                        pass

                if data:  # pragma: no cover
                    with self._threads_lock:
                        thread = threading.Thread(
                            target=_thread,
                            kwargs=dict(thread_address=address, thread_data=data),
                        )
                        thread.daemon = True
                        self.threads.append(thread)
                        thread.start()
            except Exception:
                pass
